/* 
   WebDAV XML response handling
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef DAV_XML_H
#define DAV_XML_H

/* for BUFSIZ */
#include <stdio.h>

#include <xmlparse.h>

/* WebDAV XML response handling...
   
   Basic principle is that you provide these things:
     1) a list of elements which you wish to handle 
     2) a context checking function which will validate whether a 
        given element is valid in the given parent.
     3) a callback function which is passed a completed element
    
   This code then deals with the boring stuff like:
     - Dealing with XML namespaces   
     - Collecting CDATA

   The element list is stored in a 'struct dav_xml_elm' array.
   For each element, you specify the element name, an element id,
   and some flags for that element: whether it can have children,
   and whether it can have cdata.
      
*/

/* Reserved element id's */
#define DXE_unknown -1
#define DXE_root 0

typedef int dav_xml_elmid;

struct dav_xml_elm {
    const char *name;
    dav_xml_elmid id;
    unsigned int flags;
};

struct dav_xml_state {
    const struct dav_xml_elm *elm;
    char *default_ns; /* The current default namespace */
    struct dav_xml_nspace *nspaces; /* List of other namespace scopes */
    struct dav_xml_state *parent; /* The parent in the tree */
};

/* Function to check element context... returns 0 if child is a valid
 * child tag of parent, else non-zero. */
/* TODO: do we need userdata here? theoretically, not */
typedef int (*dav_xml_check_context)( dav_xml_elmid child, dav_xml_elmid parent );

/* Called when a complete element is parsed */
typedef int (*dav_xml_element_cb)( void *userdata, 
				   const struct dav_xml_state *s, 
				   const char *cdata );

/* We pass around a dav_xmlparser as the userdata using expat.  This
 * maintains the current state of the parse and various other bits and
 * bobs. Within the parse, we store the current branch of the tree,
 * i.e., the current element and all its parents, but nothing other
 * than that. 
 */
struct dav_xmlparser {
    struct dav_xml_state *root; /* the root of the branch */
    struct dav_xml_state *current; /* current element in the branch */
    struct http_buffer *cdata; /* the CDATA buffer */
    unsigned int valid:1; /* currently valid? */
    unsigned int want_cdata:1; /* currently collection CDATA? */
    const struct dav_xml_elm *elements;
    dav_xml_check_context context_cb;
    dav_xml_element_cb element_cb;	
    void *userdata;
    XML_Parser parser;
    char error[BUFSIZ];
};

/* Flags */
/* This element has no children */
#define DAV_CDATA (1<<1)
/* Collect complete contents of this node as cdata */
#define DAV_COLLECT (1<<2)

/* Initialise the parser p, with the list of elements we accept,
 * the context checking function, the element callback, and the userdata
 * for the element callback. */
int dav_xml_init( struct dav_xmlparser *p, const struct dav_xml_elm *elements, 
		  dav_xml_check_context ctx_func, 
		  dav_xml_element_cb elm_func, void *userdata );

/* Cleans up after a parse.
 * Returns 0 if the parse was valid, non-zero otherwise */
int dav_xml_finish( struct dav_xmlparser *p );

/* Parse the given block of input of length len */
void dav_xml_parse( struct dav_xmlparser *p, const char *block, size_t len );
/* As above taking a void * argument */
void dav_xml_parse_v( void *userdata, const char *block, size_t len );


#endif /* DAV_XML_H */
