/* 
   cadaver, command-line DAV client
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   $Id: commands.c,v 1.24 1999/12/30 13:24:14 joe Exp $
*/

#include <config.h>

#include <sys/stat.h>
#include <sys/wait.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#include <errno.h>

/* readline requires FILE *, silly thing */
#include <stdio.h>

#ifdef HAVE_READLINE_H
#include <readline.h>
#elif HAVE_READLINE_READLINE_H
#include <readline/readline.h>
#endif

#ifndef HAVE_SNPRINTF
#include "snprintf.h"
#endif

#include "string_utils.h"

#include "httpdav.h"
#include "davlocks.h"

#include "dates.h"
#include "cadaver.h"
#include "basename.h"
#include "dirname.h"

#include "commands.h"

/* Local variables */
bool child_running; /* true when we have a child running */

const static struct {
    enum command_id id;
    const char *name;
} command_names[] = {
    /* The direct mappings */
#define C(n) { cmd_##n, #n }
    C(ls), C(cd), C(quit), C(open), C(close), C(set), C(unset),
    C(help), C(put), C(get), C(mkcol), C(delete), C(move), C(copy),
    C(less), C(cat), C(lpwd), C(lcd), C(lls), C(echo), C(quit),
    C(mget), C(mput), C(rmcol), C(lock), C(unlock), C(discover), C(steal),
    C(showlocks),
#undef C
    /* And now the aliases */
    { cmd_less, "more" }, { cmd_mkcol, "mkdir" }, 
    { cmd_delete, "rm" }, { cmd_copy, "cp"}, { cmd_move, "mv" }, 
    { cmd_help, "h" }, { cmd_help, "?" },
    { cmd_quit, "exit" }, { cmd_quit, "bye" },
    { cmd_unknown, NULL }
};    

struct dav_lock *lock_list = NULL;

/* The actual commands */

void init_locking( const char *lockstore ) {
    DEBUG( DEBUG_LOCKS, "init locking.\n" );
    dav_lock_list = &lock_list;
    /* TODO: read in lock list from ~/.davlocks */
}

void finish_locking( const char *lockstore ) {
    DEBUG( DEBUG_LOCKS, "Finish locking.\n" );
    /* TODO: write out lock list to ~/.davlocks */
}

/* Auxiliaries */

static void 
map_multi( void (*func)( const char * ), int argc, const char *argv[] );

void simple_copy( const char *from, const char *to );
static void simple_put( const char *local, const char *remote );
static void simple_move( const char *src, const char *dest );

static void do_copymove( int argc, const char *argv[],
		  const char *v1, const char *v2,
		  void (*cb)( const char *, const char *) );

static char *clever_path( const char *path, const char *src, 
			  const char *dest );

static char *choose_pager(void);

static void dispatch( const char *verb, const char *filename, 
		      int (*func)( const char *), const char *pass );
static void dispatch2( const char *verb, const char *fn1, const char *fn2, 
		       int (*func)( const char *, const char *) );


static void display_help_message( void );

static void print_lock( struct dav_lock *lock );

const struct command *get_command( const char *name ) {
    int n, m;
    for( n = 0; command_names[n].name != NULL; n++ ) {
	if( strcasecmp( command_names[n].name, name ) == 0 ) {
	    for( m = 0; commands[m].id != cmd_unknown; m++ ) {
		if( commands[m].id == command_names[n].id )
		    return &commands[m];
	    }
	    return NULL;
	}
    }
    return NULL;
}

void execute_ls( const char *remote ) {
    int ret;
    char *real_remote;
    struct proto_file *files, *current, *next;
    if( remote != NULL ) {
	real_remote = resolve_path( path, remote, true );
    } else {
	real_remote = strdup( path );
    }
    ret = dav_fetch( real_remote, 1, 0, &files );
    if( ret == PROTO_OK ) {
	/* Easy this, eh? */
	if( files == NULL ) {
	    printf( "Collection `%s' is empty.\n", real_remote );
	} else {
	    printf( "Listing of collection `%s':\n", real_remote );
	    for( current = files; current!=NULL; current = next ) {
		char *date = rfc1123_date( current->modtime ), *restype;
		next = current->next;
		switch( current->type ) {
		case proto_file: restype = "Resource"; break;
		case proto_link: restype = "Reference"; break;
		case proto_dir: restype = "Collection"; break;
		default: restype = "Unknown"; break;
		}
		printf( "%10s: %-20s %10d  %s\n", restype, current->filename, 
			current->size, date );
		free( date );
		free( current->filename );
		free( current );
	    }
	}
    } else {
	printf( "Could not retrieve listing for `%s':\n%s\n", 
		real_remote, http_error );
    }
    free( real_remote );
}

/* A bit like Haskell's map function... applies func to each
 * of the first argc items in argv */
static void 
map_multi( void (*func)( const char * ), int argc, const char *argv[] ) {
    int n;
    for( n = 0; n < argc; n++ ) {
	(*func)( argv[n] );
    }
}

void multi_mkcol( int argc, const char *argv[] ) {
    map_multi( execute_mkcol, argc, argv );
}

void multi_delete( int argc, const char *argv[] ) {
    map_multi( execute_delete, argc, argv );
}

void multi_rmcol( int argc, const char *argv[] ) {
    map_multi( execute_rmcol, argc, argv );
}

void multi_less( int argc, const char *argv[] ) {
    map_multi( execute_less, argc, argv );
}

void multi_cat( int argc, const char *argv[] ) {
    map_multi( execute_cat, argc, argv );
}

static void dispatch( const char *verb, const char *filename, 
		      int (*func)( const char *), const char *arg ) {
    printf( "%s `%s': ", verb, filename );
    fflush( stdout );
    if( (*func)( arg ) == PROTO_OK ) {
	printf( "succeeded.\n" );
    } else {
	printf( "failed: %s\n", http_error );
    }
}

static void dispatch2( const char *verb, const char *fn1, const char *fn2,
		       int (*func)(const char *, const char *) ) {
    printf( "%s `%s' to `%s': ", verb, fn1, fn2 );
    fflush( stdout );
    if( (*func)( fn1, fn2 ) == PROTO_OK ) {
	printf( "succeeded.\n" );
    } else {
	printf( "failed: %s\n", http_error );
    }
}

static char * getowner(void) {
    char ret[BUFSIZ];
#if 1
    /* mod_dav breaks if you submit an owner */
    return strdup("");
#else
    snprintf( ret, BUFSIZ, "<href>mailto:%s@%s</href>", 
	      getenv("USER"), getenv("HOSTNAME") );
    return strdup(ret);
#endif
}

static int has_trailing_slash( const char *uri ) {
    return (uri[strlen(uri)-1] == '/');
}

/* Resolve path, appending trailing slash if resource is a
 *  collection. */
static char *true_path( const char *res ) {
    char *full;
    full = resolve_path( path, res, false );
    if( is_collection(full) ) {
	if( !has_trailing_slash(full) ) {
	    char *tmp = malloc(strlen(full)+2);
	    strcpy( tmp, full );
	    strcat( tmp, "/" );
	    free( full );
	    full = tmp;
	}
    }
    return full;
}

static const char *get_lockscope( enum dav_lock_scope s ) {
    switch( s ) {
    case dav_lockscope_exclusive: return "exclusive";
    case dav_lockscope_shared: return "shared";
    default: return "unknown";
    }
}

static const char *get_locktype( enum dav_lock_type t ) {
    if( t == dav_locktype_write ) {
	return "write";
    } else {
	return "unknown";
    }
}

static const char *get_timeout( int t ) {
    static char buf[128];
    switch( t ) {
    case DAV_TIMEOUT_INFINITE: return "infinite";
    case DAV_TIMEOUT_INVALID: return "invalid";
    default:
	sprintf( buf, "%d", t );
	return buf;
    }
}

static char *get_depth( int d ) {
    switch( d ) {
    case DAV_DEPTH_INFINITE:
	return "infinite";
    case 0:
	return "0";
    case 1:
	return "1";
    default:
	return "invalid";
    }
}

static void print_lock( struct dav_lock *lock ) {
    printf( "Lock <%s> on %s:\n"
	    "  Scope: %s  Type: %s  Timeout: %s\n"
	    "  Owner: %s  Depth: %s\n", 
	    lock->token, lock->uri?lock->uri:"(unknown)",
	    get_lockscope( lock->scope ),
	    get_locktype( lock->type ), get_timeout( lock->timeout ),
	    lock->owner?lock->owner:"(none)", get_depth(lock->depth) );
}

void execute_discover( const char *res ) {
    struct lock_result *list, *cur;
    char *real_remote;
    real_remote = resolve_path( path, res, false );
    printf( "Discovering locks on `%s': ", res );
    fflush( stdout );
    switch( dav_lock_discover( real_remote, &list ) ) {
    case PROTO_OK:
	printf( "succeeded.\n" );
	for( cur = list; cur != NULL; cur=cur->next ) {
	    /* TODO: I don't think we can actually find out the URI which
	     * the lock applies to. */
	    print_lock( &cur->lock );
	}
	/* TODO: free the list */
	break;
    case PROTO_NONE:
	printf( "No locks found.\n" );
	break;
    default:
	printf( "failed:\n%s\n", http_error );
    }
}

void execute_steal( const char *res ) {
    struct lock_result *list, *cur;
    char *real_remote;
    real_remote = resolve_path( path, res, false );
    printf( "Stealing locks on `%s': ", res );
    fflush( stdout );
    switch( dav_lock_discover( real_remote, &list ) ) {
    case PROTO_OK:
	printf( "succeeded.\n" );
	for( cur = list; cur != NULL; cur=cur->next ) {
	    cur->lock.uri = cur->href;
	    printf( "%s: <%s>\n", cur->href, cur->lock.token );
	    dav_lock_add( &lock_list, &cur->lock );
	}
	/* TODO: free the list */
	break;
    case PROTO_NONE:
	printf( "No locks found.\n" );
	break;
    default:
	printf( "failed:\n%s\n", http_error );
    }
}

void execute_showlocks( ) {
    struct dav_lock *lock;
    if( lock_list == NULL ) {
	printf( "No owned locks.\n" );
    } else {
	printf( "List of owned locks:\n" );
	for( lock = lock_list; lock!= NULL; lock = lock->next ) {
	    print_lock( lock );
	}
    }
}

void execute_lock( const char *res ) {
    char *real_remote;
    struct dav_lock *lock;
    lock = malloc( sizeof(struct dav_lock) );
    if( !lock ) return;
    real_remote = true_path( res );
    lock->type = dav_locktype_write;
    lock->scope = dav_lockscope_exclusive;
    lock->owner = getowner();
    if( has_trailing_slash(real_remote) ) {
	lock->depth = DAV_DEPTH_INFINITE;
    } else {
	lock->depth = 0;
    }
    lock->uri = real_remote;
    printf( "Locking `%s': ", res );
    fflush( stdout );
    if( dav_lock( lock ) == PROTO_OK ) {
	printf( "succeeded.\nLocktoken: <%s>\n", lock->token );
	dav_lock_add( &lock_list, lock );
    } else {
	printf( "failed: %s\n", http_error );
	free( real_remote );
    }
}

void execute_unlock( const char *res ) {
    struct dav_lock *lock;
    char *real_remote;
    int isdummy = 0;
    real_remote = true_path( res );
    lock = dav_lock_find( lock_list, real_remote );
    if( !lock ) {
	lock = malloc( sizeof(struct dav_lock) );
	lock->token = readline( "Enter locktoken: " );
	if( !lock->token || strlen(lock->token) == 0 ) {
	    free( lock );
	    return;
	}
	lock->uri = real_remote;
	isdummy=1;
    }
    printf( "Unlocking `%s': ", res );
    if( dav_unlock( lock ) == PROTO_OK ) {
	printf( "succeeded.\n" );
    } else {
	printf( "failed: %s\n", http_error );
    }
    if( isdummy ) {
	free( lock );
    } else {
	dav_lock_remove( &lock_list, lock );
    }
}

void execute_mkcol( const char *filename ) {
    char *remote;
    remote = resolve_path( path, filename, true );
    dispatch( "Creating", filename, dav_mkcol, remote );
    free( remote );
}

void execute_delete( const char *filename ) {
    char *remote;
    remote = resolve_path( path, filename, false );
    if( is_collection( remote ) ) {
	printf( "Will not delete collection resource `%s'.\n", filename );
    } else {
	dispatch( "Deleting", filename, http_delete, remote );
    }
    free( remote );
}

void execute_rmcol( const char *filename ) {
    char *remote;
    remote = resolve_path( path, filename, true );
    if( !is_collection( remote ) ) {
	printf( "Will not delete non-collection resource `%s'.\n", filename );
    } else {
	dispatch( "Deleting", filename, http_delete, remote );
    }
    free( remote );
}

/* Like resolve_path except more intelligent. */
static char *clever_path( const char *path, const char *src, 
			  const char *dest ) {
    char *ret;
    bool src_is_coll, dest_is_coll;
    dest_is_coll = (dest[strlen(dest)-1] == '/');
    src_is_coll = (src[strlen(src)-1] == '/');
    if( strcmp( dest, "." ) == 0 ) {
	ret = resolve_path( path, base_name( src ), false );
    } else if( strcmp( dest, ".." ) == 0 ) {
	char *path2, *parent;
	/* We need the parent of path... 
	 * Sadly, dir_name("/a/b/") == dir_name("/a/b/")
	 * So, dup path and strip off the trailing slash.
	 */
	path2 = strdup( path );
	path2[strlen(path2)-1] = '\0';
	parent = dir_name( path2 );
	ret = resolve_path( parent, base_name( src ), false );
	free( parent );
	free( path2 );
    } else if( !src_is_coll && dest_is_coll ) {
	/* Moving a file to a collection... the destination should
	 * be the basename of file concated with the collection. */
	char *tmp;
	tmp = resolve_path( path, dest, true );
	CONCAT2( ret, tmp, base_name( src ) );
    } else {
	ret = resolve_path( path, dest, false );
    }
    return ret;
}

static void cat_callback( void *userdata, const char *buf, const size_t len ) {
    FILE *f = userdata;
    if( fwrite( buf, len, 1, f ) < 0 ) {
	perror( "fwrite" );
    }
}

static char *choose_pager(void) {
    struct stat st;
    char *tmp;
    tmp = getenv("PAGER");
    if( tmp != NULL ) {
	return tmp;
    } else if( stat( "/usr/bin/less", &st ) == 0 ) {
	return "/usr/bin/less";
    } else if( stat( "/bin/less", &st ) == 0 ) {
	return "/bin/less";
    } else {
	return "/bin/more";
    }
}

static FILE *spawn_pager( const char *pager ) {
    /* Create a pipe */
    return popen( pager, "w" );
}

static void kill_pager( FILE *p ) {
    /* This blocks until the pager quits. */
    pclose( p );
}

void execute_less( const char *resource ) {
    char *real_res;
    char *pager;
    FILE *p;
    real_res = resolve_path( path, resource, false );
    pager = choose_pager();
    printf( "Displaying `%s':\n", real_res );
    p = spawn_pager(pager);
    if( p == NULL ) {
	printf( "Error! Could not spawn pager `%s':\n%s\n", pager,
		strerror(errno) );
    } else {
	child_running = true;
	http_read_file( real_res, 0, cat_callback, p );
	kill_pager( p ); /* Blocks until the pager quits */
	child_running = false;
    }
}

void execute_cat( const char *resource ) {
    char *real_res = resolve_path( path, resource, false );
    printf( "Displaying `%s':\n", real_res );
    (void) http_read_file( real_res, 0, cat_callback, stdout );
    printf( "%s\n", http_error );	
}

/* Returns non-zero if given resource is not a collection resource */
int is_collection( const char *res ) {
    struct proto_file *files;
    /* the zero-length filename check here may be unnecessary */
    if (dav_fetch( res, 0, 1, &files ) == PROTO_OK &&
	files != NULL &&
	files->filename[0] == '\0' ) {
	if( files->type == proto_dir ) {
	    return 1;
	} else {
	    snprintf( http_error, BUFSIZ, "Not a collection resource." );
	    return 0;
	}
    } else {
	return 0;
    }
}

void multi_copy( int argc, const char *argv[] ) {
    do_copymove( argc, argv, "copying", "copy", simple_copy );
}

void multi_move( int argc, const char *argv[] ) {
    do_copymove( argc, argv, "moving", "move", simple_move );
}

static void do_copymove( int argc, const char *argv[],
		  const char *v1, const char *v2,
		  void (*cb)( const char *, const char *) ) {
    /* We are guaranteed that argc > 2... */
    char *dest;

    dest = resolve_path(path, argv[argc-1], true );
    if( is_collection( dest ) ) {
	int n;
	char *real_src, *real_dest;
	for( n = 0; n < argc-1; n++ ) {
	    real_src = resolve_path(path, argv[n], false );
	    real_dest = clever_path(path, argv[n], dest);
	    if( strcmp( real_src, real_dest ) == 0 ) {
		printf( "%s: %s and %s are the same resource.\n", v1,
			real_src, real_dest );
	    } else {
		(*cb)( real_src, real_dest );
	    }
	    free( real_src );
	    free( real_dest );
	}
    } else if( argc > 2 ) {
	/* nicked from fileutils */
	printf( "When %s multiple resources, last argument must be a collection.\n", v2 );
    } else {
	char *rsrc, *rdest;
	rsrc = resolve_path( path,  argv[0], false );
	rdest = resolve_path( path, argv[1], false );
	/* Simple */
	(*cb) ( rsrc, rdest );
	free( rsrc );
	free( rdest );
    }
    free( dest );
}

void simple_move( const char *src, const char *dest ) {
    dispatch2( "Moving", src, dest, dav_move );
}

void simple_copy( const char *src, const char *dest ) {
    dispatch2( "Copying", src, dest, dav_copy );
}

/* Returns absolute filename which is 'filename' relative to 
 * (absolute filename) 'path'. e.g.
 *    resolve_path( "/dav/", "asda" ) == "/dav/asda"
 *    resolve_path( "/dav/", "/asda" ) == "/asda"
 * Also removes '..' segments, e.g.
 *    resolve_path( "/dav/foobar/", "../fish" ) == "/dav/fish"
 * If isdir is true, ensures the return value has a trailing slash.
 */
char *resolve_path( const char *path, const char *filename,
			   bool isdir ) {
    char *ret, *pnt;
    if( *filename == '/' ) {
	/* It's absolute */
	ret = strdup( filename );
    } else if( strcmp( filename, "." ) == 0 ) {
	ret = strdup( path );
    } else {
	CONCAT2( ret, path, filename );
    }
    if( isdir && ret[strlen(ret)-1] != '/' ) {
	char *newret;
	CONCAT2( newret, ret, "/" );
	free( ret );
	ret = newret;
    }
    /* Sort out '..', etc... */
    do {
	pnt = strstr( ret, "/../" );
	if( pnt != NULL ) {
	    char *last;
	    /* Find the *previous* path segment, to overwrite...
	     *    /foo/../
	     *    ^- last points here */
	    if( pnt > ret ) {
		for( last = pnt-1; (last > ret) && (*last != '/'); last-- );
	    } else {
		last = ret;
	    }
	    memmove( last, pnt + 3, strlen(pnt+2) );
	} else {
	    pnt = strstr( ret, "/./" );
	    if( pnt != NULL ) {
		memmove( pnt, pnt+2, strlen(pnt+1) );
	    }
	}
    } while( pnt != NULL );
    return ret;    
}

void execute_get( const char *remote, const char *local ) {
    char *filename, *real_remote;
    real_remote = resolve_path( path, remote, false );
    if( local == NULL ) {
	struct stat st;
	/* Choose an appropriate local filename */
	if( stat( base_name(remote), &st ) == 0 ) {
	    /* File already exists... don't overwrite */
	    printf( "Enter local filename for `%s': ", real_remote );
	    filename = readline( NULL );
	    if( filename == NULL ) {
		free( real_remote );
		printf( "cancelled.\n" );
		return;
	    }
	} else {
	    filename = strdup( base_name( remote ) );
	}
    } else {
	filename = strdup( local );
    }
    /* Muck around like this because otherwise the connection status
     * messages and username/password prompts would mess up the nice
     * progress dots. */
    snprintf( transfer_msg, BUFSIZ, 
	      "Downloading `%s' to %s [", real_remote, filename );
    transfer_status = tran_starting;
    switch( http_get( filename, real_remote, 0, false ) ) {
    case PROTO_OK:
	if( transfer_status == tran_starting ) {
	    printf( transfer_msg );
	}
	printf( "] done.\n" );
	break;
    default:
	if( transfer_status == tran_going ) {
	    printf( "] f" );
	} else {
	    printf( "F" );
	}
	printf( "ailed:\n%s\n", http_error );
	break;
    }
    transfer_status = tran_none;
    free( real_remote );
    free( filename );
}

static void simple_put( const char *local, const char *remote ) {
    snprintf( transfer_msg, BUFSIZ,
	      "Uploading %s to `%s' [", local, remote );
    transfer_status = tran_starting;
    switch( http_put( local, remote, false ) ) {
    case PROTO_OK:
	if( transfer_status == tran_starting ) {
	    printf( transfer_msg );
	}
	printf( "] done.\n" );
	break;
    default:
	if( transfer_status == tran_going ) {
	    printf( "] f" );
	} else {
	    printf( "F" );
	}
	printf( "ailed:\n%s\n", http_error );
	break;
    }
    transfer_status = tran_none;
}

void execute_put( const char *local, const char *remote ) {
    char *real_remote;
    if( remote == NULL ) {
	real_remote = resolve_path( path, base_name( local ), false );
    } else {
	real_remote = resolve_path( path, remote, false );
    }
    simple_put( local, real_remote );
    free( real_remote );
}

/* this is getting too easy */
void multi_mput( int argc, const char *argv[] ) {
    for( ; argv[0] != NULL; argv++ ) {
	char *remote;
	remote = clever_path( path, argv[0], path );
	simple_put( argv[0], remote );
	free( remote );
    }    
}

void multi_mget( int argc, const char *argv[] ) {
    for( ; argv[0] != NULL; argv++ ) {
	execute_get( argv[0], NULL );
    }
}

void execute_lpwd( void ) {
    char pwd[BUFSIZ];
    if( getcwd( pwd, BUFSIZ ) == NULL ) {
	perror( "pwd" );
    } else {
	printf( "Local directory: %s\n", pwd );
    }
}

/* Using a hack, we get zero-effort multiple-argument 'lls' */
void execute_lls( int argc, char * const argv[] ) {
    int pid;
    pid = fork();
    switch( pid ) {
    case 0: /* child...  */
	/* FIXME? 'ls' might modify the arguments, but we're in a
	 * different process now, so who gives a monkeys?
	 * The implicit typecast munging might screw things up in
	 * weird and wonderful ways, though? */
	execvp( "ls", &argv[-1] );
	printf( "Error executing ls: %s\n", strerror(errno) );
	exit( -1 );
	break;
    case -1: /* Error */
	printf( "Error forking: %s\n", strerror(errno) );
	break;
    default: /* Parent */
	wait(NULL);
	break;
    }
    return;
}


void execute_lcd( const char *path ) {
    const char *real_path;
    if( path ) {
	real_path = path;
    } else {
	real_path = getenv("HOME");
	if( !real_path ) {
	    printf( "Could not determine home directory from environment.\n" );
	    return;
	}
    }
    if( chdir( real_path ) ) {
	printf( "Could not change local directory:\nchdir: %s\n",
		strerror(errno) );
    }
}


void execute_cd( const char *newpath ) {
    char *real_path;
    bool is_swap = false;
    bool is_coll;
    if( strcmp( newpath, "-" ) == 0 ) {
	is_swap = true;
	real_path = old_path;
    } else if( !path ) {
	real_path = strdup( newpath );
    } else {	
	real_path = resolve_path( path, newpath, true );
    }
    is_coll = is_collection( real_path );
    if( is_coll || tolerant ) {
	if( !is_coll ) {
	    printf( "Ignored: %s not WebDAV-enabled:\n%s\n", real_path,
		    http_error );
	}
	if( !is_swap && old_path ) free( old_path );
	old_path = path;
	path = real_path;
    } else {
	printf( "Could not access %s (not WebDAV-enabled?):\n%s\n", real_path, http_error );
	if( !is_swap ) free( real_path );
    }
}

static void display_help_message( void ) {
    int n;
    printf( "Commands: \n" );
    
    for( n = 0; commands[n].id != cmd_unknown; n++ ) {
	if( commands[n].call && commands[n].short_help ) {
	    printf( " %-26s %s\n", commands[n].call, commands[n].short_help );
	}
    }
    
    printf( "Aliases: rm=delete, mkdir=mkcol, mv=move, cp=copy, more=less, quit=exit=bye\n"
	    "Resource name equivalences: . -> current collection\n" );
    
}

void execute_help( const char *arg ) {
    if( !arg ) {
	display_help_message();
    } else {
	const struct command *cmd = get_command( arg );
	if( cmd ) {
	    if( cmd->short_help ) {
		printf( "  %s : %s\n", arg, cmd->short_help );
	    } else {
		printf( "No help message for command %s.\n", arg );
	    }
	} else {
	    printf( "Command name not known: %s\n", arg );
	}
    }
}

void execute_echo( int count, const char **args ) {
    const char **pnt;
    for( pnt = args; *pnt != NULL; pnt++ ) {
	printf( "%s ", *pnt );
    }
    putchar( '\n' );
}

