/* 
   cadaver, command-line DAV client
   Copyright (C) 1999, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
   
   $Id: options.c,v 1.7 1999/12/30 11:01:02 joe Exp $
*/

/* Options handling */

#include <config.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <stdio.h>

#include "common.h"
#include "httpdav.h"
#include "string_utils.h"

#include "cadaver.h"
#include "options.h"

static void *set_debug( void *old, const char *new );
static void *unset_debug( void *old, const char *new );
static void disp_debug( void *old );
#if 0
static void *set_lockowner( void *old, const char *new );
static void *unset_lockowner( void *old, const char *new );
static void disp_lockowner( void *old );
#endif
static struct {
    const char *name;
    void *holder;
    enum {
	opt_bool,
	opt_handled
    } type;
    void *(*set)( void *old, const char *val );
    void *(*unset)( void *old, const char *val );
    void (*display)( void *val );
    const char *help;
    const char *handle_help;
} options[] = {
#define B(x,v,h) { x, v, opt_bool, NULL, NULL, NULL, h, NULL }
    /* Booleans */
    B( "tolerant", &tolerant, "Tolerate non-WebDAV collections" ),
    B( "expect100", &http_enable_expect, "Enable use of 'Expect: 100-continue' header" ),
#undef B
    { "debug", &debug_mask, opt_handled, 
      set_debug, unset_debug, disp_debug, "Debugging options",
      "The debug value is a list of comma-separated keywords.\n"
      "Valid keywords are: socket, http, xml, httpauth, cleartext."
    }, /*
    { "lockowner", &lock_owner, opt_handled,
      set_lockowner, unset_lockowner, disp_lockowner, "Lock owner URI",
      "The lock owner is a URI."
      }, */
    { NULL, 0 }
};

static const struct {
    const char *name;
    int val;
} debug_map[] = {
    { "xml", DEBUG_XML },
    { "xmlparse", DEBUG_XMLPARSE },
    { "http", DEBUG_HTTP },
    { "socket", DEBUG_SOCKET },
    { "httpauth", DEBUG_HTTPAUTH },
    { "cleartext", DEBUG_HTTPPLAIN },
    { "files", DEBUG_FILES },
    { "locks", DEBUG_LOCKS },
    { NULL, 0 }
};

static void display_options( void ) {
    int n;
    printf( "Options:\n" );
    for( n = 0; options[n].name != NULL; n++ ) {
	int *val = (int *)options[n].holder;
	switch( options[n].type ) {
	case opt_bool:
	    printf( " %s is %s\n", options[n].name, *val?"set":"unset" );
	    break;
	case opt_handled:
	    printf( " %s is ", options[n].name );
	    options[n].display( options[n].holder );
	    printf( "\n" );
	    break;
	}
    }
}

static void *do_debug( void *old, const char *set, int setit ) {
    int *val = old, n;
    char **vs;
    if( !setit && !set ) return NULL;
    vs = split_string( set, ',', NULL, NULL );
    if( val == NULL ) { 
	/* init */
	val = malloc( sizeof(int) );
	*val = 0;
    }
    for( n = 0; vs[n] != NULL; n++ ) {
	int d, got = 0;
	for( d = 0; debug_map[d].name != NULL; d++ ) {
	    if( strcasecmp( vs[n], debug_map[d].name ) == 0 ) {
		if( setit ) {
		    *val |= debug_map[d].val ;
		} else {
		    *val &= ~debug_map[d].val;
		}
		got = 1;
	    }
	}
	if( !got ) {
	    printf( "Debug option %s unknown.\n", vs[n] );
	}
    }
    split_string_free( vs );
    return (void *)val;
}

static void *set_debug( void *old, const char *set ) {
    return do_debug( old, set, 1 );
}

static void *unset_debug( void *old, const char *s ) {
    return do_debug( old, s, 0 );
}

static void disp_debug( void *old  ) {
    int val = *(int *)old, n, flag=0;
    putchar( '{' );
    for( n = 0; debug_map[n].name != NULL; n++ ) {
	if( val & debug_map[n].val ) {
	    printf( "%s%s", flag++?",":"", debug_map[n].name );
	}
    }
    putchar( '}' );
}

void execute_set( const char *opt, const char *newv ) {
    if( opt == NULL ) {
	display_options();
    } else {
	int n;
	for( n = 0; options[n].name != NULL; n++ ) {
	    int *val = (int *)options[n].holder;
	    if( strcasecmp( options[n].name, opt ) == 0 ) {
		switch( options[n].type ) {
		case opt_bool:
		    if( newv ) {
			printf( "%s is a boolean option.\n", opt );
		    } else {
			*val = 1;
		    }
		    break;
		case opt_handled:
		    if( !newv ) {
			printf( "%s must be given a value:\n%s\n", opt,
				options[n].handle_help );
		    } else {
			options[n].holder = (*options[n].set)( options[n].holder, newv );
		    }
		    break;
		}
		return;
	    }
	}
	printf( "Unknown option: %s.\n", opt );
    }
}

void execute_unset( const char *opt, const char *newv ) {
    int n;
    for( n = 0; options[n].name != NULL; n++ ) {
	int *val = (int *)options[n].holder;
	void *newval;
	if( strcasecmp( options[n].name, opt ) == 0 ) {
	    switch( options[n].type ) {
	    case opt_bool:
		if( newv != NULL ) {
		    printf( "%s ia a boolean option.\n", opt );
		} else {
		    *val = 0;
		}
		break;
	    case opt_handled:
		newval = (*options[n].unset)( options[n].holder, newv );
		if( !newval ) {
		    printf( "%s must be given a value:\n%s\n", opt,
			    options[n].handle_help );
		} else {
		    options[n].holder = newval;
		}
		break;
	    }
	    return;
	}
    }
    printf( "Unknown option: %s.\n", opt );
}
